is_production = {
    "درآمدهاي عملياتي": "sales",
    "بهاى تمام شده درآمدهاي عملياتي": "cost_of_goods_sold",
    "سود(زيان) ناخالص": "gross_profit",
    "هزينه‏ هاى فروش، ادارى و عمومى": "marketing_general_and_administrative_expenses",
    "هزينه کاهش ارزش دريافتني‏ ها (هزينه استثنايي)": "special_items",
    "ساير درآمدها": "other_operating_income",
    "سایر درآمدهای عملیاتی": "other_operating_income",
    "سایر هزینه ها": "other_operating_expense",
    "سایر هزینه‌های عملیاتی": "other_operating_expense",
    "سود(زيان) عملياتى": "operating_income",
    "هزينه‏ هاى مالى": "interest_expense",
    "ساير درآمدها و هزينه ‏هاى غيرعملياتى": "other_income",
    "سایر درآمدها و هزینه‌های غیرعملیاتی- درآمد سرمایه‌گذاری‌ها": "other_income",
    "سایر درآمدها و هزینه‌های غیرعملیاتی- اقلام متفرقه": "other_income",
    "سود(زيان) عمليات در حال تداوم قبل از ماليات": "income_from_continuing_operations_before_taxes",
    "مالیات بر درآمد": "income_tax",
    "سال جاري": "income_tax",
    "سال‌هاي قبل": "income_tax",
    "سود(زيان) خالص عمليات در حال تداوم": "net_income_from_continuing_operations",
    "سود (زيان) خالص عمليات متوقف شده": "income_from_discontinued_operations_net_of_tax",
    "سود (زیان) عملیات متوقف ‌شده پس از اثر مالیاتی": "income_from_discontinued_operations_net_of_tax",
    "سود(زيان) خالص": "net_income",
    "سود (زيان) خالص هر سهم – ريال": "eps",
    "سرمایه": "listed_capital",
}

is_bank = {
    "درآمد تسهيلات اعطايي و سپرده گذاری و اوراق بدهی": "income_on_credit_facilities",
    "سود و وجه التزام تسهیلات اعطایی": "income_on_credit_facilities",  # o
    "سود (زیان) حاصل از سرمایه‌گذاری‌ها و سپرده‌گذاری‌ها": "income_on_credit_facilities",  # o
    "هزينه سود سپرده‌ها": "depositors_interest_share",
    "سود علی‌الحساب سپرده‌های سرمایه‌گذاری": "depositors_interest_share",  # o
    "سود علی‌الحساب سپرده‌های ": "depositors_interest_share",  # o
    "خالص درآمد تسهیلات و سپرده گذاری": "net_facility_and_deposit_income",
    "درآمد کارمزد": "commission_income",
    "هزينه کارمزد": "commission_expense",
    "خالص درآمد کارمزد": "net_commission_income",
    "خالص سود (زیان) سرمایه گذاری‌ها": "income_on_investments_and_deposits",
    "سود و وجه التزام فعالیت‌های غیرمشاع": "income_on_investments_and_deposits",  # o
    "خالص سود (زيان) مبادلات و معاملات ارزي": "foreign_exchange_transaction_gain",
    "نتیجه مبادلات ارزی": "foreign_exchange_transaction_gain",  # o
    "ساير درآمدها و هزینه های عملياتي": "other_operating_income_and_expenses",
    "سایر درآمدها": "other_operating_income_and_expenses",  # o
    "جمع درآمدهاي عملياتي": "total_operating_income",
    "خالص ساير درآمدها و هزینه‌ها": "net_other_income_and_expenses",
    "هزينه‌هاي اداري و عمومي": "corporate_and_administrative_expenses",
    "هزینه‌های کارکنان": "corporate_and_administrative_expenses",  # o
    "سایر هزینه‌های اجرایی": "corporate_and_administrative_expenses",  # o
    "هزينه مطالبات مشکوک الوصول": "doubtful_debts_expense",
    "هزينه‌هاي مالي": "finance_expense",
    "هزينه استهلاک": "depreciation_expense",
    "سود(زیان) عملیات در حال تداوم قبل از مالیات": "earning_before_tax",
    "ماليات بر درآمد سال جاری": "tax",
    "ماليات بر درآمد سال های قبل": "tax",
    "مالیات بر درآمد": "tax",  # o
    "سود (زيان) خالص عملیات در حال تداوم": "net_income_from_continuing_operation",
    "سود(زیان) خالص عملیات متوقف شده": "discontinued_operations_profit",
    "سود (زیان) عملیات متوقف ‌شده پس از اثر مالیاتی": "discontinued_operations_profit",  # o
    "سود(زیان) خالص": "net_income",
    "سود (زیان) خالص هر سهم– ریال": "eps",
    "سرمایه": "listed_capital",
}

is_insurance = {
    "درآمد حق بیمه ناخالص",
    "premium_earned",
    "حق بیمه اتکایی واگذاری",
    "expense_of_reinsurance_transfer_premium",
    "حق بیمه خالص (سهم نگهداری)",
    "premium_earned_retention",
    "درآمد سرمایه گذاری از محل منابع بیمه ای",
    "investment_income_from_technical_reserves",
    "سایر درآمدهای بیمه ای",
    "other_insurance_income",
    "درآمدهای بیمه ای",
    "insurance_income",
    "خسارت و مزایای پرداختی (ناخالص)",
    "cost_of_claims",
    "خسارت سهم بیمه گران اتکایی",
    "cost_of_claims_reinsurers_hare",
    "خسارت و مزایای پرداختی خالص",
    "cost_of_Claims_Retention",
    "هزینه سهم مشارکت در منافع",
    "cost_of_interest_participating",
    "تغییرات سایر ذخایر فنی",
    "other_technical_reserves_changes",
    "سایر هزینه های بیمه ای",
    "other_insurance_costs",
    "هزینه های بیمه ای",
    "insurance_costs",
    "سود (زیان) ناخالص فعالیتهای بیمه ای",
    "gross_profit_of_insurance_activities",
    "هزینه های اداری و عمومی",
    "sales_general_and_administrative_expense",
    "سایر درآمدها و هزینه های عملیاتی",
    "other_operating_income_and_expenses",
    "سود (زیان) عملیاتی",
    "operating_profit",
    "هزینه های مالی",
    "finance_expense",
    "سایر درآمدها و هزینه های غیر عملیاتی",
    "net_other_non_operating_gains_and_expenses",
    "سود (زیان) قبل از مالیات",
    "earning_before_tax",
    "مالیات بر درآمد",
    "tax",
    "سود (زیان) خالص",
    "net_income",
    "سود (زیان) خالص هر سهم– ریال",
    "eps",
    "سرمایه",
    "listed_capital",
}
